#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2020 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np

from .baseclass import MatcherBaseclass


class HistogramMatcher(MatcherBaseclass):

    def __init__(self, *args, **kwargs):
        super(HistogramMatcher, self).__init__(*args, **kwargs)

    def hist_match(self, src: np.ndarray=None, ref: np.ndarray=None) -> np.ndarray:
        '''

        This function conducts channel-wise histogram matching which is invariant of image resolutions,
        but requires the same number of color channels in both images.

        :param src: Source image that requires transfer
        :param ref: Palette image which serves as reference
        :param ref: Resulting image after the mapping

        :type src: :class:`~numpy:numpy.ndarray`
        :type ref: :class:`~numpy:numpy.ndarray`
        :type result: :class:`~numpy:numpy.ndarray`

        :return: **result**
        :rtype: np.ndarray

        '''

        # override source and reference image with arguments (if provided)
        self._src = src if src is not None else self._src
        self._ref = ref if ref is not None else self._ref

        # parameter init
        result = np.zeros_like(self._src)

        for ch in range(self._src.shape[2]):

            # convert to 1D arrays
            src_vec = self._src[..., ch].ravel()
            ref_vec = self._ref[..., ch].ravel()

            # analyze histograms
            _, src_idxs, src_cnts = np.unique(src_vec, return_inverse=True, return_counts=True)
            ref_vals, ref_cnts = np.unique(ref_vec, return_counts=True)

            # compute cumulative distribution functions
            src_cdf = np.cumsum(src_cnts).astype(np.float64) / src_vec.size
            ref_cdf = np.cumsum(ref_cnts).astype(np.float64) / ref_vec.size

            # do the histogram mapping
            interp_vals = np.interp(src_cdf, ref_cdf, ref_vals)
            result[..., ch] = interp_vals[src_idxs].reshape(src[..., ch].shape)

        return result



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "inbox@christopherhahne.de"
__license__ = """
Copyright (c) 2020 Christopher Hahne <inbox@christopherhahne.de>
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__version__ = '0.1.6'

from .top_level import ColorMatcher
from .hist_matcher import HistogramMatcher
from .mvgd_matcher import TransferMVGD
from .baseclass import MatcherBaseclass



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2020 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np

from .baseclass import MatcherBaseclass
from types import FunctionType


class TransferMVGD(MatcherBaseclass):

    def __init__(self, *args, **kwargs):
        super(TransferMVGD, self).__init__(*args, **kwargs)

        self._fun = kwargs['fun'] if 'fun' in kwargs else self.mkl_solver

    def transfer(self, src: np.ndarray=None, ref: np.ndarray=None, fun: FunctionType=None) -> np.ndarray:
        '''

        Transfer function to map colors based on for Multi-Variate Gaussian Distributions (MVGDs).

        :param src: Source image that requires transfer
        :param ref: Palette image which serves as reference
        :param fun: optional argument to pass a transfer function to solve for covariance matrices
        :param t_r: Resulting image after the mapping

        :type src: :class:`~numpy:numpy.ndarray`
        :type ref: :class:`~numpy:numpy.ndarray`
        :type t_r: :class:`~numpy:numpy.ndarray`

        :return: **t_r**
        :rtype: np.ndarray

        '''

        # check if three color channels are provided
        self.validate_color_chs()

        # set solver function for transfer matrix (default is MKL)
        self._fun = fun if fun is not None else self._fun

        r = np.reshape(src, [-1, src.shape[2]])
        z = np.reshape(ref, [-1, ref.shape[2]])

        cov_r = np.cov(r.T)
        cov_z = np.cov(z.T)

        transfer_mat = self._fun(cov_r, cov_z)

        mu_r = np.mean(r, axis=0)
        mu_z = np.mean(z, axis=0)

        t_r = np.dot((r - mu_r), transfer_mat) + mu_z
        t_r = np.reshape(t_r, src.shape)

        return t_r

    @staticmethod
    def mkl_solver(cov_r: np.ndarray, cov_z: np.ndarray):
        '''

        This function computes the transfer matrix based on the Monge-Kantorovich linearization.

        :param cov_r: Covariance matrix of source image
        :param cov_z: Covariance matrix of reference image
        :param T: Transfer matrix

        :type cov_r: :class:`~numpy:numpy.ndarray`
        :type cov_z: :class:`~numpy:numpy.ndarray`
        :type T: :class:`~numpy:numpy.ndarray`

        :return: **T**
        :rtype: np.ndarray

        '''

        [Da2, Ua] = np.linalg.eig(cov_r)
        Ua = np.array([Ua[:, 2] * -1, Ua[:, 1], Ua[:, 0] * -1]).T
        Da2[Da2 < 0] = 0
        Da = np.diag(np.sqrt(Da2[::-1]))
        C = np.dot(Da, np.dot(Ua.T, np.dot(cov_z, np.dot(Ua, Da))))
        [Dc2, Uc] = np.linalg.eig(C)
        Dc2[Dc2 < 0] = 0
        Dc = np.diag(np.sqrt(Dc2))
        Da_inv = np.diag(1. / (np.diag(Da + np.spacing(1))))

        T = np.dot(Ua, np.dot(Da_inv, np.dot(Uc, np.dot(Dc, np.dot(Uc.T, np.dot(Da_inv, Ua.T))))))

        return T



from color_matcher import ColorMatcher
from color_matcher.io_handler import load_img_file
from color_matcher.normalizer import Normalizer

import os
import numpy as np
import imageio
try:
    from PIL import Image, ImageSequence
except:
    raise Exception('Please install pillow')

dir_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data')
fn_img1 = "wave.gif"
if fn_img1.endswith('gif'):
    im = Image.open(os.path.join(dir_path, fn_img1))
duration = 40  #im.info['duration'] if 'duration' in im.info else
img2 = load_img_file(os.path.join(dir_path, 'sunrise_mvgd.png'))

sequence = []
method = 'mvgd'
#size = 200, 200
for frame in ImageSequence.Iterator(im):

    #img1 = np.asarray(frame.convert('RGB').resize(size, Image.ANTIALIAS), np.uint8)
    img1 = np.asarray(frame.convert('RGB'), np.uint8)

    # create color match object
    match = ColorMatcher(img1, img2, method=method).main()

    sequence.append(Image.fromarray(Normalizer(match).uint8_norm()))

output_fn = os.path.join(dir_path, os.path.splitext(fn_img1)[0] + '_' + method + '.gif')
imageio.mimwrite(output_fn, sequence, duration=1/duration, palettesize=2**8)



import requests
import os
from tqdm import tqdm
from urllib.parse import urlparse

try:
    from bs4 import BeautifulSoup as bs
except ImportError:
    raise Exception('Please install bs4 package')


def is_valid(url):
    """
    Checks whether `url` is a valid URL.
    """
    parsed = urlparse(url)
    return bool(parsed.netloc) and bool(parsed.scheme)


def get_all_images(url):
    """
    Returns all image URLs on a single `url`
    """

    soup = bs(requests.get(url).content, "html.parser")

    filenames = [x.attrs.get('href') for x in soup.find_all("a") if x.attrs.get('href').__contains__('bmp')]
    urls = [os.path.join(url, filename) for filename in filenames]

    return urls


def download_file(url, pathname):
    """
    Downloads a file given an URL and puts it in the folder `pathname`
    """
    # if path doesn't exist, make that path dir
    if not os.path.isdir(pathname):
        os.makedirs(pathname)
    # download the body of response by chunk, not immediately
    response = requests.get(url, stream=True)
    # get the total file size
    file_size = int(response.headers.get("Content-Length", 0))
    # get the file name
    filename = os.path.join(pathname, url.split("/")[-1])
    # progress bar, changing the unit to bytes instead of iteration (default by tqdm)
    progress = tqdm(response.iter_content(1024), f"Downloading {filename}", total=file_size, unit="B", unit_scale=True, unit_divisor=1024)
    with open(filename, "wb") as f:
        for data in progress:
            # write data read to the file
            f.write(data)
            # update the progress bar manually
            progress.update(len(data))


def download_stack(url, path):
    # get all images
    imgs = get_all_images(url)
    for img in imgs:
        # for each image, download it
        download_file(img, path)


if __name__ == '__main__':
    download_stack('https://www.math.purdue.edu/~lucier/PHOTO_CD/BMP_IMAGES/', os.path.join(os.getcwd(), 'data'))



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "inbox@christopherhahne.de"
__license__ = """
Copyright (c) 2020 Christopher Hahne <inbox@christopherhahne.de>
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np


class MatcherBaseclass(object):

    def __init__(self, *args, **kwargs):

        self._src = None
        self._ref = None

        if len(args) == 2:
            self._src = args[0]
            self._ref = args[1]

        if bool(kwargs):
            self._src = kwargs['src'] if 'src' in kwargs else self._src
            self._ref = kwargs['ref'] if 'ref' in kwargs else self._ref

        self.validate_img_dims()

    def validate_img_dims(self):
        '''
        This function validates the image dimensions. It throws an exception if the dimension are unequal to 2 or 3.
        '''

        # add third image dimension for monochromatic images
        self._src = self._src[..., np.newaxis] if len(self._src.shape) == 2 else self._src
        self._ref = self._ref[..., np.newaxis] if len(self._ref.shape) == 2 else self._ref

        if len(self._src.shape) != 3 or len(self._ref.shape) != 3:
            raise BaseException('Wrong image dimensions')

    def validate_color_chs(self):
        '''
        This function checks whether provided images consist of 3 color channels. An exception is thrown otherwise.
        '''

        if self._src.shape[2] != 3 or self._ref.shape[2] != 3:
            raise BaseException('Each image must have 3 color channels')



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2020 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import os
import numpy as np

try:
    from PIL import Image
except ImportError:
    raise ImportError('Please install pillow.')

from color_matcher.normalizer import Normalizer

FILE_EXTS = ('bmp', 'png', 'tiff', 'tif', 'jpeg', 'jpg')


def save_img_file(img, file_path=None, file_type=None):

    file_path = os.getcwd() if file_path is None else file_path
    ext = os.path.splitext(file_path)[-1][1:]

    if not file_type:
        file_type = ext if ext == 'png' or ext == 'tiff' else 'tiff' if img.dtype == 'uint16' else 'png'

    # compose new file path string if extension type changed
    file_path = os.path.splitext(file_path)[-2] if file_path.endswith(FILE_EXTS) else file_path
    file_type = 'png' if file_type is None else file_type
    file_path += '.' + file_type

    # normalization
    img = Normalizer(img).uint16_norm() if file_type.__contains__('tif') else Normalizer(img).uint8_norm()

    try:
        import imageio
        suppress_user_warning(True, category=UserWarning)
        imageio.imwrite(uri=file_path, im=img)
        suppress_user_warning(False, category=UserWarning)
    except ImportError:
        if file_type == 'png' or file_type == 'bmp':
            try:
                Image.fromarray(img).save(file_path, file_type, optimize=True)
                #imageio.imwrite(uri=file_path, im=img)
            except PermissionError as e:
                raise Exception(e)

    return True


def load_img_file(file_path):

    # get file extension
    file_type = file_path.split('.')[-1]

    if any(file_type.lower() in ext for ext in FILE_EXTS):
        try:
            import imageio
            suppress_user_warning(True, category=UserWarning)
            img = imageio.imread(uri=file_path, format=file_type)
            suppress_user_warning(False, category=UserWarning)
        except ImportError:
            try:
                img = Image.open(file_path)
            except OSError or TypeError:
                # support load of truncated images
                from PIL import ImageFile
                ImageFile.LOAD_TRUNCATED_IMAGES = True
                img = Image.open(file_path)

    else:
        raise TypeError('Filetype %s not recognized' % file_type)

    # normalize (convert to numpy array)
    img = Normalizer(np.asarray(img)).type_norm()

    return img


def suppress_user_warning(switch=None, category=None):

    import warnings
    switch = switch if switch is None else True
    if switch:
        warnings.filterwarnings("ignore", category=category)
    else:
        warnings.filterwarnings("default", category=category)


def select_file(init_dir=None, title=''):
    ''' get filepath from tkinter dialog '''

    # consider initial directory if provided
    init_dir = os.path.expanduser('~/') if not init_dir else init_dir

    # import tkinter while considering Python version
    try:
        import tkinter as tk
        from tkinter.filedialog import askopenfilename
    except ImportError:
        import Tkinter as tk
        from tkFileDialog import askopenfilename

    # open window using tkinter
    root = tk.Tk()
    root.withdraw()
    root.update()
    file_path = askopenfilename(initialdir=[init_dir], title=title)
    root.update()

    return file_path if file_path else None


#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "inbox@christopherhahne.de"
__license__ = """
    Copyright (c) 2020 Christopher Hahne <inbox@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from color_matcher.top_level import ColorMatcher
from color_matcher.io_handler import *
from .img_downloader import download_stack

import unittest
import os


class MatchKodakTester(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(MatchKodakTester, self).__init__(*args, **kwargs)

    def setUp(self):

        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        self.dat_path = os.path.join(self.dir_path, 'data')


    def test_kodak_images(self):
        # prepare data
        url = 'https://www.math.purdue.edu/~lucier/PHOTO_CD/BMP_IMAGES/'
        self.fnames = ['IMG' + str(i + 1).zfill(4) + '.bmp' for i in range(24)]
        loc_path = os.path.join(self.dat_path, 'kodak')

        try:
            os.makedirs(loc_path, 0o755)
            os.makedirs(os.path.join(loc_path, 'results'), 0o755)
        except:
            pass

        if not os.path.exists(loc_path):
            download_stack(url, loc_path)

        for fn_img1 in self.fnames:
            for fn_img2 in self.fnames:
                # load images
                img1 = load_img_file(os.path.join(loc_path, fn_img1))
                img2 = load_img_file(os.path.join(loc_path, fn_img2))

                # create color match object
                res = ColorMatcher(img1, img2, method='hm-mkl-hm').main()

                # assess quality
                val = self.avg_hist_dist(res, img2)
                print('Avg. histogram distance %s' % val)

                # save result
                output_filename = os.path.join(loc_path, 'results', fn_img1.split('.')[0] + '_from_' + fn_img2)
                save_img_file(res, file_path=output_filename)

                # assertion
                self.assertEqual(True, val != 0)



import numpy as np

from .baseclass import MatcherBaseclass
from color_matcher.normalizer import Normalizer

# Observer. = 2°, Illuminant = D65
REF_X = 95.047
REF_Y = 100.000
REF_Z = 108.883


class ReinhardMatcher(MatcherBaseclass):

    def __init__(self, *args, **kwargs):
        super(ReinhardMatcher, self).__init__(*args, **kwargs)

    def reinhard(self, src: np.ndarray=None, ref: np.ndarray=None) -> np.ndarray:
        '''

        This function conducts color matching in Lab color space based on the principles proposed by Reinhard et al.
        The paper of the original work can be found at https://www.cs.tau.ac.il/~turkel/imagepapers/ColorTransfer.pdf

        :param src: Source image that requires transfer
        :param ref: Palette image which serves as reference
        :param ref: Resulting image after the mapping

        :type src: :class:`~numpy:numpy.ndarray`
        :type ref: :class:`~numpy:numpy.ndarray`
        :type res: :class:`~numpy:numpy.ndarray`

        :return: **result**
        :rtype: np.ndarray

        '''

        # convert to Lab color space
        src_lab = self.rgb2lab(src)
        ref_lab = self.rgb2lab(ref)

        # compute color statistics for the source and reference images
        src_means, src_stds = self.gauss_analysis(src_lab)
        ref_means, ref_stds = self.gauss_analysis(ref_lab)
        res = (src_stds/ref_stds) * (src_lab-src_means) + ref_means

        # convert to RGB space
        res = self.lab2rgb(res)
        res = Normalizer(res).uint8_norm()

        # return the color transferred image
        return res

    def gauss_analysis(self, img):

        # compute the mean and standard deviation of each channel
        l, a, b = np.dsplit(img, 3)
        l_mean, l_std = l.mean(), l.std()
        a_mean, a_std = a.mean(), a.std()
        b_mean, b_std = b.mean(), b.std()

        # return the color statistics
        return np.array([l_mean, a_mean, b_mean]), np.array([l_std, a_std, b_std])

    def rgb2xyz(self, rgb):
        '''
        https://web.archive.org/web/20120502065620/http://cookbooks.adobe.com/post_Useful_color_equations__RGB_to_LAB_converter-14227.html
        '''

        rgb = rgb / np.max(rgb)

        for ch in range(rgb.shape[2]):
            mask = rgb[..., ch] > 0.04045
            rgb[..., ch][mask] = np.power((rgb[..., ch] + 0.055) / 1.055, 2.4)[mask]
            rgb[..., ch][~mask] /= 12.92

        rgb *= 100

        # Observer. = 2°, Illuminant = D65 (from Adobe)
        mat_adb = np.array([[0.4124, 0.2126, 0.0193], [0.3576, 0.7152, 0.1192], [0.1805, 0.0722, 0.9505]])

        # from Reinhard et al. paper (2001)
        mat_itu = np.array([[0.4306, 0.2220, 0.0202], [0.3415, 0.7067, 0.1295], [0.1784, 0.0713, 0.9394]])

        xyz = np.dot(rgb, mat_adb)

        return xyz

    def xyz2lab(self, xyz):

        xyz[..., 0] /= REF_X
        xyz[..., 1] /= REF_Y
        xyz[..., 2] /= REF_Z

        for ch in range(xyz.shape[2]):
            mask = xyz[..., ch]>0.008856
            xyz[..., ch][mask] = np.power(xyz[..., ch], 1/3.)[mask]
            xyz[..., ch][~mask] = (7.787*xyz[..., ch] + 16/116.)[~mask]

        lab = np.zeros(xyz.shape)
        lab[..., 0] = (116 * xyz[..., 1]) - 16
        lab[..., 1] = 500 * (xyz[..., 0]-xyz[..., 1])
        lab[..., 2] = 200 * (xyz[..., 1]-xyz[..., 2])

        return lab

    def lab2xyz(self, lab):

        xyz = np.zeros(lab.shape)
        xyz[..., 1] = (lab[..., 0] + 16) / 116.
        xyz[..., 0] = lab[..., 1] / 500. + xyz[..., 1]
        xyz[..., 2] = xyz[..., 1] - lab[..., 2] / 200.

        for ch in range(xyz.shape[2]):
            mask = np.power(xyz[..., ch], 3) > 0.008856
            xyz[..., ch][mask] = np.power(xyz[..., ch], 3)[mask]
            xyz[..., ch][~mask] = (xyz[..., ch] - 16/116.)[~mask] / 7.787

        xyz[..., 0] *= REF_X
        xyz[..., 1] *= REF_Y
        xyz[..., 2] *= REF_Z

        return xyz

    def xyz2rgb(self, xyz):

        xyz /= 100.

        # Observer. = 2°, Illuminant = D65
        mat = np.array([[3.2406, -0.9689, -0.0557], [-1.5372,  1.8758, -0.204], [-0.4986, -0.0415,  1.057]])
        rgb = np.dot(xyz, mat)

        for ch in range(rgb.shape[2]):
            mask = rgb[..., ch] > 0.0031308
            rgb[..., ch][mask] = 1.055 * np.power(rgb[..., ch][mask], 1/2.4) - 0.055
            rgb[..., ch][~mask] *= 12.92

        return rgb

    def rgb2lab(self, rgb):

        xyz = self.rgb2xyz(rgb)
        lab = self.xyz2lab(xyz)

        return lab

    def lab2rgb(self, lab):
        xyz = self.lab2xyz(lab)
        rgb = self.xyz2rgb(xyz)

        return rgb



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2020 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from color_matcher.hist_matcher import HistogramMatcher
from color_matcher.mvgd_matcher import TransferMVGD
from color_matcher.reinhard_matcher import ReinhardMatcher
import numpy as np

METHODS = ('default', 'mvgd', 'hm', 'hm-mkl-hm', 'reinhard')


class ColorMatcher(HistogramMatcher, TransferMVGD, ReinhardMatcher):

    def __init__(self, *args, **kwargs):
        super(ColorMatcher, self).__init__(*args, **kwargs)

        self._method = kwargs['method'] if 'method' in kwargs else 'default'

    def main(self, method: str=None) -> np.ndarray:
        '''
        The main function and high-level entry point performing the mapping. Valid methods are

        :param method: ('default', 'mvgd', 'hm', 'hm-mkl-hm') describing how to conduct color mapping
        :type method: :class:`str`

        :return: Resulting image after color mapping
        :rtype: np.ndarray
        '''

        self._method = self._method if method is None else method

        # color transfer methods (to be iterated through)
        if self._method == METHODS[0]:
            funs = [self.transfer]
        elif self._method == METHODS[1]:
            funs = [self.transfer]
        elif self._method == METHODS[2]:
            funs = [self.hist_match]
        elif self._method == METHODS[3]:
            funs = [self.hist_match, self.transfer, self.hist_match]
        elif self._method == METHODS[4]:
            funs = [self.reinhard]
        else:
            raise BaseException('Method type \'%s\' not recognized' % method)

        # proceed with the color match
        for fun in funs:
            self._src = fun(self._src, self._ref)

        return self._src



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2020 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from color_matcher import __version__
from color_matcher.top_level import ColorMatcher, METHODS
from color_matcher.io_handler import *

import getopt
import sys, os


def usage():

    print("Usage: color-matcher <options>\n")
    print("Options:")
    print("-s <path>,     --src=<path>       Specify source image file or folder to process")
    print("-r <filepath>, --ref=<filepath>   Specify target image file")
    print("-m <method>,   --method=<method>  Provide color transfer method. Available methods are:")
    print("                                  "+', '.join(['"'+m+'"' for m in METHODS]))
    print("-w ,           --win              Select files from window")
    print("-h,            --help             Print this help message")
    print("")


def parse_options(argv):

    try:
        opts, args = getopt.getopt(argv, "hs:r:m:w", ["help", "src=", "ref=", "method=", "win"])
    except getopt.GetoptError as e:
        print(e)
        sys.exit(2)

    cfg = dict()

    # default settings (use test data images for MKL conversion)
    cfg['src_path'] = ''
    cfg['ref_path'] = ''
    cfg['method'] = None
    cfg['win'] = None

    if opts:
        for (opt, arg) in opts:
            if opt in ("-h", "--help"):
                usage()
                sys.exit()
            if opt in ("-s", "--src"):
                cfg['src_path'] = arg
            if opt in ("-r", "--ref"):
                cfg['ref_path'] = arg
            if opt in ("-m", "--method"):
                cfg['method'] = arg
            if opt in ("-w", "--win"):
                cfg['win'] = True

    # create dictionary containing all parameters for the light field
    return cfg


def main():

    # program info
    print("\ncolor-matcher v%s \n" % __version__)

    # parse options
    cfg = parse_options(sys.argv[1:])

    # select files from window (if option set)
    if cfg['win']:
        cfg['src_path'] = select_file('.', 'Select source image')
        cfg['ref_path'] = select_file(cfg['src_path'], 'Select reference image')

    # cancel if file paths not provided
    if not cfg['src_path'] or not cfg['ref_path']:
        usage()
        print('Canceled due to missing image file path\n')
        sys.exit()

    # select light field image(s) considering provided folder or file
    if os.path.isdir(cfg['src_path']):
        filenames = [f for f in os.listdir(cfg['src_path']) if f.lower().endswith(FILE_EXTS)]
    elif not os.path.isfile(cfg['src_path']) or not os.path.isfile(cfg['ref_path']):
        print('File(s) not found \n')
        sys.exit()
    else:
        filenames = [cfg['src_path']]


    # method handling
    cfg['method'] = cfg['method'] if cfg['method'] in METHODS else 'mvgd'

    # file handling
    ref = load_img_file(cfg['ref_path'])
    output_path = os.path.dirname(cfg['src_path'])

    # process the images
    for f in filenames:
        src = load_img_file(f)
        res = ColorMatcher(src=src, ref=ref, method=cfg['method']).main()
        filename = os.path.splitext(os.path.basename(cfg['src_path']))[0]+'_'+cfg['method']
        file_ext = os.path.splitext(cfg['src_path'])[-1]
        save_img_file(res, file_path=os.path.join(output_path, filename), file_type=file_ext[1:])

    return True


if __name__ == "__main__":

    sys.exit(main())



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "inbox@christopherhahne.de"
__license__ = """
    Copyright (c) 2020 Christopher Hahne <inbox@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from color_matcher.top_level import ColorMatcher, METHODS
from color_matcher.io_handler import *

import unittest
import os
import numpy as np
from ddt import ddt, idata, unpack

@ddt
class MatchMethodTester(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(MatchMethodTester, self).__init__(*args, **kwargs)

    def setUp(self):

        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        self.dat_path = os.path.join(self.dir_path, 'data')

    @staticmethod
    def avg_hist_dist(img1, img2, bins=2**8-1):

        hist_a = np.histogram(img1, bins)[0]
        hist_b = np.histogram(img2, bins)[0]

        return np.sqrt(np.sum(np.square(hist_a - hist_b)))

    @idata(([m] for m in METHODS))
    @unpack
    def test_match_method(self, method=None):

        # skip if no method present
        if method is None:
            self.skipTest('Type \'None\' was passed and skipped')

        # load images
        plain = load_img_file(os.path.join(self.dat_path, 'scotland_plain.png'))
        house = load_img_file(os.path.join(self.dat_path, 'scotland_house.png'))
        refer = load_img_file(os.path.join(self.dat_path, 'scotland_pitie.png'))

        # create color match object
        match = ColorMatcher(src=house, ref=plain, method=method).main()

        # assess quality
        refer_val = self.avg_hist_dist(plain, refer)
        match_val = self.avg_hist_dist(plain, match)
        print('\nAvg. histogram distance of original %s vs. %s %s' % (round(refer_val, 3), method, round(match_val, 3)))

        # assertion
        self.assertEqual(True, refer_val > match_val)

    def test_cli(self):

        from color_matcher.bin.cli import main
        import sys

        # compose cli arguments
        sys.argv.append('--src='+os.path.join(self.dat_path, 'scotland_house.png'))
        sys.argv.append('--ref='+os.path.join(self.dat_path, 'scotland_plain.png'))

        # run cli command
        ret = main()

        # assertion
        self.assertEqual(True, ret)

    @unittest.skipUnless(False, "n.a.")
    def test_match_method_imageio(self, method=None):

        # get test data from imageio lib
        import imageio
        fn_img1 = 'chelsea'
        fn_img2 = 'coffee'
        img1 = imageio.imread('imageio:'+fn_img1+'.png')
        img2 = imageio.imread('imageio:'+fn_img2+'.png')

        # create color match object
        match = ColorMatcher(img1, img2, method=method).main()

        # assess quality
        match_val = self.avg_hist_dist(match, img2)
        print('Avg. histogram distances %s vs %s' % (float('inf'), match_val))

        # save result
        loc_path = './test/data'
        output_filename = os.path.join(loc_path, fn_img1.split('.')[0] + '_from_' + fn_img2)
        save_img_file(img1, file_path=os.path.join(loc_path, fn_img1))
        save_img_file(img2, file_path=os.path.join(loc_path, fn_img2))
        save_img_file(match, file_path=output_filename)

        # assertion
        self.assertEqual(True, float('inf') > match_val)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np


class Normalizer(object):

    def __init__(self, data=None, min=None, max=None):

        self._data, self._min, self._max = None, None, None
        self._var_init(data, min, max)

    def _var_init(self, data=None, min=None, max=None):

        self._data = self._data if data is None else np.asarray(data, dtype='float64')
        self._dtype = str(self._data.dtype) if isinstance(self._data, np.ndarray) else 'float64'

        self._min = self._min if min is None else min
        self._max = self._max if max is None else max
        self._min = self._data.min() if not any([self._min, min]) and isinstance(self._data, np.ndarray) else self._min
        self._max = self._data.max() if not any([self._max, max]) and isinstance(self._data, np.ndarray) else self._max

    def uint16_norm(self):
        ''' normalize image array to 16-bit unsigned integer '''

        return np.asarray(np.round(self.norm_fun()*(2**16-1)), dtype=np.uint16)

    def uint8_norm(self):
        ''' normalize image array to 8-bit unsigned integer '''

        return np.asarray(np.round(self.norm_fun()*(2**8-1)), dtype=np.uint8)

    def type_norm(self, data=None, min=None, max=None, new_min=None, new_max=None):
        ''' normalize numpy image array for provided data type '''

        self._var_init(data, min, max)

        if self._dtype.startswith(('int', 'uint')):
            new_max = np.iinfo(np.dtype(self._dtype)).max if new_max is None else new_max
            new_min = np.iinfo(np.dtype(self._dtype)).min if new_min is None else new_min
            img_norm = np.round(self.norm_fun() * (new_max - new_min) + new_min)
        else:
            new_max = 1.0 if new_max is None else new_max
            new_min = 0.0 if new_min is None else new_min
            img_norm = self.norm_fun() * (new_max - new_min) + new_min

        return np.asarray(img_norm, dtype=self._dtype)

    def norm_fun(self):
        ''' normalize image to values between 1 and 0 '''

        norm = (self._data - self._min) / (self._max - self._min) if self._max != (self._min and 0) else self._data

        # prevent wrap-around
        norm[norm < 0] = 0
        norm[norm > 1] = 1

        return norm
